grammar Waifu2;

WS: ' ';
NL: '\n';

ID
    : (Letter | '_') (Letter | '_' | DEC_DIGIT)* 
    ;

INDENT: '  ';

fragment Letter: [a-zA-Z] ;

// comments
MultiLineComment
    : '##' ( MultiLineComment | . )*? '##'
    ;

SingleLineComment
    : '#' ~[\u000A]*
    ;
 
// operators
MULT: '*' ;
MOD: '%' ;
DIV: '/' ;
ADD: '+' ;
SUB: '-' ;

ASSIGN: '<-';

// RESERVED

LPAREN: '(' ;
RPAREN: ')' ;
COLON: ':' ;

HASH: '#';

// numerics
fragment DEC_DIGIT:
    [0-9]
    ;
INT: DEC_DIGIT+ ;
DEC: DEC_DIGIT* '.' DEC_DIGIT+ ;
BOOL: [01] ;

STR: '\'' .*? '\'' ;
TYPE: COLON .+? ;

num: INT | DEC | BOOL;

sign
    : ADD
    | SUB
    ;
    
value: DEC | INT | BOOL | STR ;

operator: SUB | ADD | MULT | DIV | MOD ;

literal: INT | DEC | BOOL | STR ;

// func
func: functionDeclaration NL functionBody (NL* | EOF);

functionDeclaration: LPAREN functionParam* RPAREN (WS TYPE)? WS ID WS*;
functionParam: ID WS TYPE ;

functionBody: (INDENT expr) (WS* NL INDENT expr)*;
functionCall: ID WS expr (WS expr)* ;

emptyExpr: WS*;

// expr
expr: literal
    | ID
    | DEC
    | expr WS operator WS expr
    | SUB (ID | num)
    | ID WS ASSIGN WS expr
    | MultiLineComment
    | SingleLineComment
    | functionCall
    | emptyExpr
    ;
 
waifuFile
    : (WS | NL)* (expr NL)* (WS | NL)* func*
    ;
